/* this keyboard interface is based on avr appnote 313
   changes are made for using avr-gcc and atmega8 processors
   beck@informatik.htw-dresden.de
*/

#include <avr/pgmspace.h>
#include "kb.h"
#include "uart.h"
#include "gpr.h"
#include "pindefs.h"
#define BUFF_SIZE 64

void decode(unsigned char sc);
void put_kbbuff(unsigned char c);

volatile unsigned char edge, bitcount;                // 0 = neg.  1 = pos.

unsigned char kb_buffer[BUFF_SIZE];
volatile unsigned char *inpt, *outpt;
volatile unsigned char buffcnt;


void init_kb(void)
{
    cli();
    inpt =  kb_buffer;                        // Initialize buffer
    outpt = kb_buffer;
    buffcnt = 0;
    //GIMSK=(1<<INT0);
    GIMSK |= (1<<INT0); //Interrupt an INTx aktivieren
    //MCUCR = 2;                                // INT0 interrupt on falling edge
    MCUCR &= ~(1<<ISC00);
    MCUCR |= (1<<ISC01);
    edge = 0;                                // 0 = falling edge  1 = rising edge
    bitcount = 11;
    sei();
}

ISR(INT0_vect)
{
    static unsigned char data;                // Holds the received scan code

    if(!edge)                                 // Routine entered at falling edge
    {
        if(bitcount < 11 && bitcount > 2)    // Bit 3 to 10 is data. Parity bit,
        {
            // start and stop bits are ignored.
            data = (data >> 1);

            if((PORT_KB & (1<<DATAPIN)))
            {
                data = data | 0x80;            // Store a '1'
            }
        }

        //MCUCR = 3;                            // Set interrupt on rising edge
        MCUCR |= (1<<ISC00);
        MCUCR |= (1<<ISC01);
        edge = 1;
    }
    else
    {
        // Routine entered at rising edge
        //MCUCR = 2;                            // Set interrupt on falling edge
        MCUCR &= ~(1<<ISC00);
        MCUCR |= (1<<ISC01);
        edge = 0;
        bitcount-=1;

        if(bitcount == 0)                    // All bits received
        {
            decode(data);
            bitcount = 11;
        }
    }
}

void decode(unsigned char sc)
{
// Coding us-keyboard, geman keyboard contains multi-byte chars
// Vector ist indexed by scancode and contains the ascii code at
// the indexed position
    static char CodingUnshifted[]=
    {
        /*  0   1   2  3   4   5   6  7   8   9   a   b   c   d   e   f*/
        /*0*/	 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,'^' ,0   ,
        /*1*/	 0   ,0   ,0   ,0   ,0   ,'q' ,'1' ,0   ,0   ,0   ,'y' ,'s' ,'a' ,'w' ,'2' ,0   ,
        /*2*/	 0   ,'c' ,'x' ,'d' ,'e' ,'4' ,'3' ,0   ,0   ,' ' ,'v' ,'f' ,'t' ,'r' ,'5' ,0   ,
        /*3*/	 0   ,'n' ,'b' ,'h' ,'g' ,'z' ,'6' ,0   ,0   ,',' ,'m' ,'j' ,'u' ,'7' ,'8' ,0   ,
        /*4*/	 0   ,',' ,'k' ,'i' ,'o' ,'0' ,'9' ,0   ,0   ,'.' ,'-' ,'l' ,'' ,'p' ,'' ,0   ,
        /*5*/	 0   ,0   ,'',0   ,'' ,0xb4,0   ,0   ,0   ,0   ,13  ,'+' ,0   ,'#', 0   ,0   ,
        /*6*/	 0   ,'<' ,0   ,0   ,0   ,0   ,8   ,0   ,0   ,'1' ,0   ,'4' ,'7' ,0   ,0   ,0   ,
        /*7*/	 '0' ,',' ,'2' ,'5' ,'6' ,'8' ,0x1b   ,0   ,0   ,'+' ,'3' ,'-' ,'*' ,'9' ,0   ,0
    };
    static char CodingCtrl[]=
    {
        /*0*/	 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,9   ,0   ,0   ,
        /*1*/	 0   ,0   ,0   ,0   ,4   ,'Q' ,'!' ,0   ,0   ,0   ,'Y' ,'S' ,'A' ,'W' ,'@' ,0   ,
        /*2*/	 0   ,3 ,'X' ,'D' ,'E' ,'$' ,'#' ,0   ,0   ,' ' ,'V' ,'F' ,'T' ,'R' ,'%' ,0   ,
        /*3*/	 0   ,'N' ,'B' ,'H' ,'G' ,'Z' ,'^' ,0   ,0   ,',' ,'M' ,'J' ,'U' ,'&' ,'*' ,0   ,
        /*4*/	 0   ,'<' ,'K' ,'I' ,'O' ,')' ,'(' ,0   ,0   ,'>' ,'/' ,'L' ,';' ,'P' ,'_' ,0   ,
        /*5*/	 0   ,0   ,'"' ,0   ,'{' ,'=' ,0   ,0   ,0   ,0   ,13  ,'}' ,0   ,'\\',0   ,0   ,
        /*6*/	 0   ,'>' ,0   ,0   ,0   ,0   ,8   ,0   ,0   ,'1' ,0   ,'4' ,'7' ,0   ,0   ,0   ,
        /*7*/	 '0' ,';' ,'2' ,'5' ,'6' ,'8' ,0x1b   ,0   ,0   ,'+' ,'3' ,'-' ,'*' ,'9' ,0   ,0
    };
    static char CodingAltGr[]=
    {
        /*0*/	 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,9   ,0   ,0   ,
        /*1*/	 0   ,0   ,0   ,0   ,4   ,'@' ,'' ,0   ,0   ,0   ,'Y' ,'S' ,'A' ,'W' ,'' ,0   ,
        /*2*/	 0   ,'' ,'X' ,'D' ,'' ,'' ,'' ,0   ,0   ,' ' ,'V' ,'F' ,'T' ,'R' ,'%' ,0   ,
        /*3*/	 0   ,'N' ,'B' ,'H' ,'G' ,'Z' ,'' ,0   ,0   ,',' ,'' ,'J' ,'' ,'{' ,'[' ,0   ,
        /*4*/	 0   ,'<' ,'K' ,'I' ,'O' ,'}' ,']' ,0   ,0   ,'>' ,'/' ,'L' ,';' ,'P' ,'\\' ,0   ,
        /*5*/	 0   ,0   ,'"' ,0   ,'{' ,'' ,0   ,0   ,0   ,0   ,13  ,'~' ,0   ,'\\',0   ,0   ,
        /*6*/	 0   ,'|' ,0   ,0   ,0   ,0   ,8   ,0   ,0   ,'1' ,0   ,'4' ,'7' ,0   ,0   ,0   ,
        /*7*/	 '0' ,';' ,'2' ,'5' ,'6' ,'8' ,0   ,0   ,0   ,'+' ,'3' ,'-' ,'*' ,'9' ,0   ,0
    };
    static char CodingShifted[]=
    {
        /*0*/	 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,9   ,''   ,0   ,
        /*1*/	 0   ,0   ,0   ,0   ,4   ,'Q' ,'!' ,0   ,0   ,0   ,'Y' ,'S' ,'A' ,'W' ,'"' ,0   ,
        /*2*/	 0   ,'C' ,'X' ,'D' ,'E' ,'$' ,'' ,0   ,0   ,' ' ,'V' ,'F' ,'T' ,'R' ,'%' ,0   ,
        /*3*/	 0   ,'N' ,'B' ,'H' ,'G' ,'Z' ,'&' ,0   ,0   ,',' ,'M' ,'J' ,'U' ,'/' ,'(' ,0   ,
        /*4*/	 0   ,';' ,'K' ,'I' ,'O' ,'=' ,')' ,0   ,0   ,':' ,'_' ,'L' ,'' ,'P' ,'?' ,0   ,
        /*5*/	 0   ,0   ,'' ,0   ,'' ,'`' ,0   ,0   ,0   ,0   ,13  ,'*' ,0   ,'\'',0   ,0   ,
        /*6*/	 0   ,'>' ,0   ,0   ,0   ,0   ,8   ,0   ,0   ,'1' ,0   ,'4' ,'7' ,0   ,0   ,0   ,
        /*7*/	 '0' ,';' ,'2' ,'5' ,'6' ,'8' ,0   ,0   ,0   ,'+' ,'3' ,'-' ,'*' ,'9' ,0   ,0
    };
    static unsigned char is_up=0, shift = 0, mode = 0;

    if(!is_up)                 // Last data received was the up-key identifier
    {
        switch(sc)
        {
        case 0xF0 :        // The up-key identifier
            is_up = 1;
            break;

        case 0x12 :        // Left SHIFT
            shift = 1;
            break;

        case 0x59 :        // Right SHIFT
            shift = 1;
            break;

        case 0x14 :        // Left CTRL
            shift = 2;
            break;

        case 0xE0 :        // AltGr
            shift = 3;
            break;

        case 0x05 :        // F1
            if(mode == 0)
            {
                mode = 1;    // Enter scan code mode
            }

            if(mode == 2)
            {
                mode = 3;    // Leave scan code mode
            }

            break;

        case 0x75 :        // UP
            put_kbbuff(0x1b); // ESC
            put_kbbuff('A');
            break;

        case 0x72 :        // DOWN
            put_kbbuff(0x1b); // ESC
            put_kbbuff('B');
            break;

        case 0x6b :        // LEFT
            put_kbbuff(0x1b); // ESC
            put_kbbuff('D');
            break;

        case 0x74 :        // RIGHT
            put_kbbuff(0x1b);
            put_kbbuff('C');
            break;

        default
                :
            if((mode == 0) || (mode == 3))        // If ASCII mode
            {
                if(sc<128)
                {
                    switch(shift)
                    {
                    case 0 :	// unshiftet
                        put_kbbuff(CodingUnshifted[sc]);
                        break;

                    case 1 :	// shiftet
                        put_kbbuff(CodingShifted[sc]);
                        break;

                    case 2 :	// ctrl
                        put_kbbuff(CodingCtrl[sc]);
                        break;

                    case 3 :	// ctrl
                        put_kbbuff(CodingAltGr[sc]);
                        break;
                    }

                    //if(!shift)                    // If shift not pressed,
                    //      put_kbbuff(CodingUnshifted[sc]);
                    //else
                    //        put_kbbuff(CodingShifted[sc]);
                }
            }
            else
                // Scan code mode
            {
                print_hexbyte(sc);            // Print scan code
                put_kbbuff(' ');
                put_kbbuff(' ');
            }

            break;
        }
    }
    else
    {
        is_up = 0;                            // Two 0xF0 in a row not allowed

        switch(sc)
        {
        case 0x12 :                        // Left SHIFT
            shift = 0;
            break;

        case 0x59 :                        // Right SHIFT
            shift = 0;
            break;

        case 0x14 :                        // Left CTRL
            shift = 0;
            break;

        case 0x11 :                        // AltGr
            shift = 0;
            break;

        case 0x05 :                        // F1
            if(mode == 1)
            {
                mode = 2;
            }

            if(mode == 3)
            {
                mode = 0;
            }

            break;

        case 0x06 :                        // F2
            //clr();
            break;
        }
    }
}

void put_kbbuff(unsigned char c)
{
    if(buffcnt<BUFF_SIZE)                         // If buffer not full
    {
        *inpt = c;                                // Put character into buffer
        inpt++;                                    // Increment pointer
        buffcnt++;

        if(inpt >= kb_buffer + BUFF_SIZE)         // Pointer wrapping
        {
            inpt = kb_buffer;
        }
    }
}

int getchar()
{
    int byte;

    //while(buffcnt == 0)
    //    ;                        // Wait for data
    if(buffcnt == 0)		// buffer empty
        return 256;                        // return bit8=1

    byte = *outpt;                                // Get byte
    outpt++;                                    // Increment pointer

    if(outpt >= kb_buffer + BUFF_SIZE)             // Pointer wrapping
    {
        outpt = kb_buffer;
    }

    buffcnt--;                                    // Decrement buffer count
    return byte;
}

void getkb()
{
    int c;

    //while(count<n)
    //while(1)
    //{
        c=getchar();
	if(c&256)   // bit8=1 => no data
	    return;
	else
            putchar((char)c);// Echo to rs232 or lcd
    //}

    //return pBuf;
}
