#include <avr/pgmspace.h>
#include <avr/io.h>
#include "uart.h"
#include "gpr.h"
#include <util/delay.h>
#include "pindefs.h"
#include "ms.h"
#define BUFF_SIZE 64
#define MDATA 5
#define MCLK 3

volatile unsigned char edge, bitcount;                // 0 = neg.  1 = pos.

unsigned char ms_buffer[BUFF_SIZE];
volatile unsigned char *msinpt, *msoutpt;
volatile unsigned char msbuffcnt;

void gohi(int pin)
{
    switch(pin)
    {
    case MDATA:
	DDRD &= ~(1<<MS_DATAPIN_DIR);
	PORTD |= (1<<MS_DATAPIN);
        break;

    case MCLK:
	DDRD &= ~(1<<MS_CLOCK_DIR);
	PORTD |= (1<<MS_CLOCK);
        break;
    };
}

void golo(int pin)
{
    switch(pin)
    {
    case MDATA:
	PORTD &= ~(1<<MS_DATAPIN);
	DDRD |= (1<<MS_DATAPIN_DIR);
        break;

    case MCLK:
	PORTD &= ~(1<<MS_CLOCK);
	DDRD |= (1<<MS_CLOCK_DIR);
        break;
    };
}

void mouse_write(char data)
{
    char i;
    char parity = 1;
    /* put pins in output mode */
    gohi(MDATA);
    gohi(MCLK);
    _delay_us(200);
    golo(MCLK);
    _delay_us(200);
    golo(MDATA);
    //_delay_us(20);
    /* start bit */
    gohi(MCLK);

    /* wait for mouse to take control of clock); */
    while((PORT_MS & (1<<MS_RD_CLOCK))) ;

    /* clock is low, and we are clear to send data */
    for(i=0; i < 8; i++)
    {
        if(data & 0x01)
        {
            gohi(MDATA);
        }
        else
        {
            golo(MDATA);
        }

        /* wait for clock cycle */
        while((PORT_MS & (1<<MS_RD_CLOCK))==0) ;

        while((PORT_MS & (1<<MS_RD_CLOCK))) ;

        parity = parity ^ (data & 0x01);
        data = data >> 1;
    }

    /* parity */
    if(parity)
    {
        gohi(MDATA);
    }
    else
    {
        golo(MDATA);
    }

    while((PORT_MS & (1<<MS_RD_CLOCK))==0) ;

    while((PORT_MS & (1<<MS_RD_CLOCK))) ;


    /* stop bit */
    gohi(MDATA);
    _delay_us(50);

    while((PORT_MS & (1<<MS_RD_CLOCK))) ;

    /* wait for mouse to switch modes */
    while(((PORT_MS & (1<<MS_RD_CLOCK))==0) || ((PORT_MS & (1<<MS_RD_DATAPIN)))) ;

    /* put a hold on the incoming data. */
    golo(MCLK);
}

signed char mouse_read(void)
{
    signed  char data = 0x00;
    int i;
    char mask = 0x01;
    /* start the clock */
    gohi(MCLK);
    gohi(MDATA);
    _delay_us(50);

    while((PORT_MS & (1<<MS_RD_CLOCK))) ;

    _delay_us(5);  /* not sure why */

    while((PORT_MS & (1<<MS_RD_CLOCK))==0) ;

    for(i=0; i < 8; i++)
    {
        while((PORT_MS & (1<<MS_RD_CLOCK))) ;

        if((PORT_MS & (1<<MS_RD_DATAPIN)))
        {
            data = data | mask;
        }

        while((PORT_MS & (1<<MS_RD_CLOCK))==0) ;

        mask = mask << 1;
    }

    /* eat parity bit, which we ignore */
    while((PORT_MS & (1<<MS_RD_CLOCK))) ;

    while((PORT_MS & (1<<MS_RD_CLOCK))==0) ;

    /* eat stop bit */
    while((PORT_MS & (1<<MS_RD_CLOCK))) ;

    while((PORT_MS & (1<<MS_RD_CLOCK))==0) ;

    /* put a hold on the incoming data. */
    golo(MCLK);
    return data;
}

void put_msbuff(unsigned char c)
{
    if(msbuffcnt<BUFF_SIZE)                         // If buffer not full
    {
        *msinpt = c;                                // Put character into buffer
        msinpt++;                                    // Increment pointer
        msbuffcnt++;

        if(msinpt >= ms_buffer + BUFF_SIZE)         // Pointer wrapping
        {
            msinpt = ms_buffer;
        }
    }
}

void mouse_init()
{
    cli();
    gohi(MCLK);
    gohi(MDATA);
    mouse_write(0xff);
    print_hexbyte(mouse_read());  /* ack byte 0xfa */
    print_hexbyte(mouse_read());  /* selftest 0xaa */
    print_hexbyte(mouse_read());  /* mouse id 0x00 */
    mouse_write(0xf4);  /* stream mode */
    print_hexbyte(mouse_read());  /* ack */
    _delay_us(100);
    gohi(MCLK);
    gohi(MDATA);

    msinpt =  ms_buffer;                        // Initialize buffer
    msoutpt = ms_buffer;
    msbuffcnt = 0;
    GIMSK |= (1<<INT1); //activate INT1
    //MCUCR = 2;                                // INT1 interrupt on falling edge
    MCUCR &= ~(1<<ISC10);
    MCUCR |= (1<<ISC11);
    edge = 0;                                // 0 = falling edge  1 = rising edge
    bitcount = 11;
    sei();
}

ISR(INT1_vect)
{
    static unsigned char data;                // Holds the received scan code

    if(!edge)                                 // Routine entered at falling edge
    {
        if(bitcount < 11 && bitcount > 2)    // Bit 3 to 10 is data. Parity bit,
        {
            // start and stop bits are ignored.
            data = (data >> 1);

            if((PORT_MS & (1<<MS_RD_DATAPIN)))
            {
                data = data | 0x80;            // Store a '1'
            }
        }

        //MCUCR = 3;                            // Set interrupt on rising edge
        MCUCR |= (1<<ISC10);
        MCUCR |= (1<<ISC11);
        edge = 1;
    }
    else
    {
        // Routine entered at rising edge
        //MCUCR = 2;                            // Set interrupt on falling edge
        MCUCR &= ~(1<<ISC10);
        MCUCR |= (1<<ISC11);
        edge = 0;
        bitcount-=1;

        if(bitcount == 0)                    // All bits received
        {
            put_msbuff(data);
            //print_hexbyte(data);            // Print scan code
            //putchar(':');            // Print scan code
            bitcount = 11;
        }
    }
}

int msgetchar()
{
    int byte;

    //while(buffcnt == 0)
    //    ;                        // Wait for data
    if(msbuffcnt == 0)		// buffer empty
        return 256;                        // return bit8=1

    byte = *msoutpt;                                // Get byte
    msoutpt++;                                    // Increment pointer

    if(msoutpt >= ms_buffer + BUFF_SIZE)             // Pointer wrapping
    {
        msoutpt = ms_buffer;
    }

    msbuffcnt--;                                    // Decrement buffer count
    return byte;
}

void getms()
{
    int c;
    static char msdata[3];
    static int datapos=0;
    int count;

        c=msgetchar(); // get byte from mouse buffer
	if(c&256)   // bit8=1 => no data
	    return;
	else
	{
	    msdata[datapos++]=c;
	    if(datapos==3)
	    {
    cli();
               putchar('1');// Send magic mouse byte - 1 for debug
	       for(count=0;count<3;count++) // send mouse packet one by one
	       {
                 putchar(msdata[count]);// Echo to rs232
               }
	       datapos=0;
               putchar('2');// Send "2" for debug
               putchar('M');// Send "M" for debug
	    sei();
	    }

        }

}
